/*
 * Project: XMON - An X protocol monitor
 *
 * File: recording.c
 *
 * Description: Routines to record and playback events to an X client
 *
 */

#ifdef RECORD_EVENTS

#include "common.h"

#include "xmond.h"
#include "linkl.h"
#include "commands.h"

/* function prototypes: */
/* recording.c: */

/* end function prototypes */

typedef struct
{
    int			wait_for_ImageText8;
    int			wait_for_PolyText8;
    unsigned char	data[32];
}
    SaveEvent;

char *event_name[] =
{
    "",		/* error packet */
    "",		/* reply packet */
    "KP",	/* KeyPress */
    "KR",	/* KeyRelease */
    "BP",	/* ButtonPress */
    "BR",	/* ButtonRelease */
    "MN",	/* MotionNotify */
    "EN",	/* EnterNotify */
    "LN",	/* LeaveNotify */
    "FI",	/* FocusIn */
    "FO",	/* FocusOut */
};

static int			ImageText8_count = 0;
static int			PolyText8_count = 0;
static LinkList			*event_list = NULL;

#define IsMouseEvent(Event)						    \
    (									    \
	Event == 4 /*ButtonPress*/ || Event == 5 /*ButtonRelease*/	    \
	||								    \
	Event == 6 /*MotionNotify*/ || Event == 7 /*EnterNotify*/	    \
	||								    \
	Event == 8 /*LeaveNotify*/					    \
    )

#define IsKeyboardEvent(Event)						    \
    (									    \
	Event == 2 /*KeyPress*/ || Event == 3 /*KeyRelease*/		    \
	||								    \
	Event == 9 /*FocusIn*/ || Event == 10 /*FocusOut*/		    \
    )

extern int			EventVerbose;

Global void
RecordEvent(server, buf, n)
    Server		    *server;
    unsigned char	    *buf;
    long		    n;
{
    short		    Event = buf[0] & 0x7f;
    int			    i;

    if (IsKeyboardEvent(Event) || IsMouseEvent(Event))
    {
	printf
	(
	    "%s %s %d %d ", PLAY_EVENT_STR, event_name[buf[0]],
	    ImageText8_count, PolyText8_count 
	);
	for(i = 0; i < 32; i++)
	{
	    putchar('A' + ((buf[i] >> 4) & 0xf));
	    putchar('A' + (buf[i] & 0xf));

	}
	putchar('\n');
	fflush(stdout);
    }
}

#ifdef USE_WRITE_NOT_PRINTF
Global void
RecordEvent(server, buf, n)
    Server		    *server;
    unsigned char	    *buf;
    long		    n;
{
    short		    Event = buf[0] & 0x7f;
    int			    i;
    int			    fd = fileno(stdout);
    int			    BytesToWrite;
    int			    num_written;
    unsigned char	    outbuf[1000]; /* enough for one line of output */
    unsigned char	    *ptr;

    if (IsKeyboardEvent(Event) || IsMouseEvent(Event))
    {
	sprintf
	(
	    outbuf, "%s %s %d %d ", PLAY_EVENT_STR, event_name[buf[0]],
	    ImageText8_count, PolyText8_count 
	);
	ptr = outbuf + strlen(outbuf);
	for(i = 0; i < 32; i++)
	{
	    *ptr++ = 'A' + ((buf[i] >> 4) & 0xf);
	    *ptr++ = 'A' + (buf[i] & 0xf);
	}
	*ptr++ = '\n';
	*ptr++ = '\0';
	/* need non-buffering output, so can't use fprint */
	BytesToWrite = strlen(outbuf);
	ptr = outbuf;
	fflush(stdout);
	do
	{
	    num_written = write(fd, ptr, BytesToWrite);
	    if (num_written > 0)
	    {
		BytesToWrite -= num_written;
		ptr += num_written;
	    }
	}
	    while (BytesToWrite > 0 && num_written > 0);
    }
}
#endif

Global void
ReplayEvent(client_list, buf)
    unsigned char	    *buf;
    LinkList		    *client_list;
{
    int			    i;
    Client		    *client;
    Server		    *server;
    Buffer		    *outbuffer;
    unsigned char	    outbuf[32];
    short		    Event;
    unsigned char	    *ptr = buf;
    int			    wait_for_ImageText8;
    int			    wait_for_PolyText8;
    SaveEvent		    *save_event;

    client = (Client *)TopOfList(client_list);
    server = (Server *)TopOfList(&client->server_list);
    outbuffer = &client->fdd->outBuffer;

    /* skip spaces */
    while (*ptr == ' ')
	ptr++;
    /* skip event name */
    while (*ptr != ' ')
	ptr++;
    /* skip spaces */
    while (*ptr == ' ')
	ptr++;
    /* read number of ImageText8 requests to wait for */
    wait_for_ImageText8 = atoi(ptr);
    /* skip ImageText8 count */
    while (*ptr != ' ')
	ptr++;
    /* skip spaces */
    while (*ptr == ' ')
	ptr++;
    /* read number of PolyText8 requests to wait for */
    wait_for_PolyText8 = atoi(ptr);
    /* skip PolyText8 count */
    while (*ptr != ' ')
	ptr++;
    /* skip spaces */
    while (*ptr == ' ')
	ptr++;

    for(i = 0; i < 32; i++)
	outbuf[i] = (((ptr[2*i] - 'A') << 4) & 0xf0) | (ptr[2*i + 1] - 'A');

    if 
    (
	(wait_for_ImageText8 <= ImageText8_count)
	&&
	(wait_for_PolyText8 <= PolyText8_count)
    )
    {
	ShortToBuf(server->SequenceNumber, &outbuf[2]);
	DecodeEvent(server, outbuf, 32, True);
	SaveBytes(outbuffer, outbuf, 32);
    }
    else
    {
	save_event = Tmalloc(SaveEvent);
	save_event->wait_for_ImageText8 = wait_for_ImageText8;
	save_event->wait_for_PolyText8 = wait_for_PolyText8;
	memcpy(save_event->data, outbuf, 32);
	if (event_list == NULL)
	{
	    event_list = Tmalloc(LinkList);
	    initList(event_list);
	}
	appendToList(event_list, (Pointer) save_event);
    }
}

Global void
RecordRequest(Request, client, buf, n)
    short		    Request;
    Client		    *client;
    unsigned char	    *buf;
    long		    n;
{
    Server		    *server;
    Buffer		    *outbuffer;
    unsigned char	    *outbuf;
    SaveEvent		    *save_event;

    if (Request == 76) /* ImageText8 */
	ImageText8_count++;
    if (Request == 74) /* PolyText8 */
	PolyText8_count++;

    if (event_list == NULL)
	return;

    server = (Server *)TopOfList(&client->server_list);
    outbuffer = &client->fdd->outBuffer;

    event_list->current = event_list->top;
    while(event_list->current != (LinkLeaf *)event_list)
    {
	save_event = (SaveEvent *)CurrentContentsOfList(event_list);
	if 
	(
	    (save_event->wait_for_ImageText8 <= ImageText8_count)
	    &&
	    (save_event->wait_for_PolyText8 <= PolyText8_count)
	)
	{
	    outbuf = save_event->data;
	    ShortToBuf(server->SequenceNumber, &outbuf[2]);
	    DecodeEvent(server, outbuf, 32, True);
	    SaveBytes(outbuffer, outbuf, 32);
	    freeCurrent(event_list);
	}
	else
	    break;
    }
}

/*
 * This was to be the beginning of a client-side X event record/playback
 * program. If you finish it let me know.
 *    - gregm@otc.otca.oz.au
 */
Global void
test_send_event(client_list)
    LinkList		    *client_list;
{
    char		    buf[32];
    Client		    *client;

    client = (Client *)TopOfList(client_list);

    buf[0] = 2;
    buf[1] = 61;
    ShortToBuf(client->SequenceNumber, &buf[2]);
    LongToBuf(0, &buf[4]);
    LongToBuf(0x0008006d, &buf[8]);
    LongToBuf(0x01100003, &buf[12]);
    LongToBuf(0, &buf[16]);
    ShortToBuf(399, &buf[20]);
    ShortToBuf(333, &buf[22]);
    ShortToBuf(78, &buf[24]);
    ShortToBuf(12, &buf[26]);
    ShortToBuf(0, &buf[28]);
    buf[30] = 1;
    buf[31] = 0;
    write (client->fdd->fd, buf, 32);
}
#endif /* RECORD_EVENTS */
