package report_window;

use strict;

##############################################################################
# report_window package - routines to display summary reports 
# from log contents.
#
# exports &create
#
# copyright (c) 1997 by Steve Romig and The Ohio State University
#
##############################################################################
# $Id: report_window.pm,v 1.3 1998/03/12 00:51:50 romig Exp $
##############################################################################
# $Log: report_window.pm,v $
# Revision 1.3  1998/03/12 00:51:50  romig
# use main::xyzzy to construct system path names, since they can
# now be relative to log file pathnames to allow separation of cases..
#
# Revision 1.1  1997/05/27 21:03:47  romig
# Initial revision
#

sub insert_report {
    my($box, $file, $title, $i) = @_;
    my($ref, $rest);

    return($i) if ! -f $file || -z $file;

    $box->insert('end', "  " . $title . "\n", $i++);

    open(REPORT, $file) ||
	warn "cannot open web report ($!)\n";

    while (<REPORT>) {
	($ref, $rest) = split(/\t/, $_, 2);

	$box->insert('end', "    ");
	$box->insert('end', $rest, ['go', "go-$ref"]);
	
	# fixme: turn ref into a clickable tag
	$box->tag('bind', $i, '<Enter>' => [
		   sub {
		       shift->tag('configure', shift, 
				  -background => $main::selection_color,
				  -relief => 'raised',
				  -border => 1);
		   }, $i]);

        $box->tag('bind', $i, '<Leave>' => [
                   sub {
		       shift->tag('configure', shift, 
				  -background => undef,
				  -relief => 'flat',
				  -border => 0);
		   }, $i]);

        $i++
    }
    
    close(REPORT);
    return($i);
}

sub redraw_reports {
    my($box, $which, @files) = @_;
    my $file;
    my $report;
    my $extension;
    my $i = 0;

    # make it writable, delete everything
    #
    $box->configure(-state => 'normal');
    $box->delete('1.0', 'end');

    foreach $file (@files) {
	&report::make_and_save($file, 0, &report::split_and_check('all'));

	my $base_dir = &main::xyzzy($main::report_dir, $file, '');

	$box->insert('end', $file. "\n", $i++);

	if ($which eq 'security' || $which eq 'all') {
	    foreach $report (@report::security_list) {
		$extension = $report::info{$report}{'extension'};
	
		$i = &insert_report($box, 
				    "$base_dir.$extension",
				    $report::info{$report}{'title'}, 
				    $i);
	    }
	}

	if ($which eq 'summary' || $which eq 'all') {
	    foreach $report (@report::summary_list) {
		$extension = $report::info{$report}{'extension'};
	
		$i = &insert_report($box, 
				    "$base_dir.$extension",
				    $report::info{$report}{'title'},
				    $i);
	    }

	    foreach $report (@report::other_list) {
		$extension = $report::info{$report}{'extension'};
	
		$i = &insert_report($box, 
				    "$base_dir.$extension",
				    $report::info{$report}{'title'},
				    $i);
	    }

	}
    }

    $box->configure(-state => 'disabled',
		    -width => 100);
}

#
# Create a scrollable window displaying the reports for a given log
# file.
#
sub create {
    my($which, @files) = @_;
    my $file;

    my $window = $log_window::main->Toplevel;

    $window->title("Reports for @files");

    # create the text widget now, i think i need it for the references
    #
    my $text = $window->Text(-relief => 'sunken',
			     -background => $main::background_color,
			     -foreground => $main::foreground_color,
			     -width => 80,
			     -height => 30,
			     -setgrid => 'yes');

    &window::define_tags($text, \&summary_window::select_traffic);

    # button bar
    #
    # +------------------------------
    # | Dismiss Redraw
    # |
    #
    my $button_bar = $window->Frame(-relief => 'raised', -bd => 2);
    $button_bar->pack(-side => 'top', -fill => 'x');

      my $dismiss_button = 
	  $button_bar->Button(-text => "Dismiss",
			   -command => ['destroy', $window]);;
      $dismiss_button->pack(-side => 'left');

      my $redraw_button = 
	  $button_bar->Button(-text => "Redraw",
			      -command => 
			          sub {&redraw_reports($text, $which, @files)});
      $redraw_button->pack(-side => 'left');

    &redraw_reports($text, $which, @files);

    # create the scroll bar
    #
    # +------------------------------
    # | Dismiss Redraw
    # + +-----------------------+ S
    # | | summary text here     | C
    # | |                       | R
    # | |                       | O
    # | |                       | L
    # | +-----------------------+ L
    #
    my $scroll = $window->Scrollbar(-command => ['yview', $text]);

    $text->configure(-yscrollcommand => ['set', $scroll]);
    $text->pack(-side => 'left',
                -fill => 'both',
		-expand => 'yes');

    $scroll->pack(-side => 'right',
		  -fill => 'y');
}

1;
