package services;

use strict;

##############################################################################
# summary package - miscellaneous routines for handling host/port pairs
# and mapping port numbers to service names and groups.
#
# exports &split_ip_port, &read_extra, &get_type, &get_tags
#
# copyright (c) 1997 by Steve Romig and The Ohio State University
#
##############################################################################
# $Id: services.pm,v 1.3 1998/03/12 00:55:54 romig Exp $
##############################################################################
# $Log: services.pm,v $
# Revision 1.3  1998/03/12 00:55:54  romig
# *** empty log message ***
#
# Revision 1.1  1997/05/27 21:03:47  romig
# Initial revision
#

# map port numbers to service names.  these *should* be read from (or
# augmented with) something like /etc/services, and *should* indicate 
# whether they are tcp or udp services.  FIXME
#
# fixme: more efficient lookups, rather than constructing 
# proto/host/port string?

%services::port2name = (
				   'tcp/*/21'	=> 'ftp',
				   'tcp/*/22'	=> 'ssh',
				   'tcp/*/23'	=> 'telnet',
				   'tcp/*/25'	=> 'smtp',
				   'tcp/*/43'	=> 'whois',
				   'tcp/*/53'	=> 'dns',
				   'tcp/*/69'	=> 'tftp',
				   'tcp/*/79'	=> 'finger',
				   'tcp/*/80'	=> 'http',
				   'tcp/*/109'	=> 'pop2',
				   'tcp/*/110'	=> 'pop',
				   'tcp/*/111'	=> 'sunrpc',
				   'tcp/*/113'	=> 'identd',
				   'tcp/*/119'	=> 'nntp',
				   'tcp/*/123'	=> 'ntp',
				   'tcp/*/137'	=> 'netbios-ns',
				   'tcp/*/138'	=> 'netbios-dgm',
				   'tcp/*/139'	=> 'netbios-ssn',
				   'tcp/*/143'	=> 'imap',
				   'tcp/*/443'	=> 'https',
				   'tcp/*/512'	=> 'rexec',
				   'tcp/*/513'	=> 'rlogin',
				   'tcp/*/514'	=> 'rsh',
				   'tcp/*/6000'	=> 'x',
				   'tcp/*/6660' => 'irc',
				   'tcp/*/6661' => 'irc',
				   'tcp/*/6662' => 'irc',
				   'tcp/*/6663' => 'irc',
				   'tcp/*/6664' => 'irc',
				   'tcp/*/6665' => 'irc',
				   'tcp/*/6666' => 'irc',
				   'tcp/*/6667' => 'irc',
				   'tcp/*/6668' => 'irc',
				   'tcp/*/6669' => 'irc',
				   'tcp/*/8080'	=> 'http',
				   );

# map services to service groups.  rlogin is a service, login is a group 
# of similar services that contains rlogin, telnet, rsh and so on.
#
%services::group = (
		   'dns'	=> '',
		   'finger'	=> '',
		   'ftp'	=> 'login',
		   'http'	=> 'web',
		   'https'	=> 'web',
		   'identd'	=> 'junk',
		   'imap'	=> 'mail',
		   'irc'	=> 'login',
		   'netbios-dgm' => '',
		   'netbios-ns'	=> '',
		   'netbios-ssn' => '',
		   'nntp'	=> 'news',
		   'ntp'	=> 'junk',
		   'pop'	=> 'mail',
		   'pop2'	=> 'mail',
		   'rexec'	=> 'login',
		   'rlogin'	=> 'login',
		   'rsh'	=> 'login',
		   'smtp'	=> 'mail',
		   'ssh'	=> 'login',
		   'sunrpc'	=> '',
		   'telnet'	=> 'login',
		   'tftp'	=> 'login',
		   'whois'	=> '',
		   'x'		=> 'login',
		   );	

#
# split strings of the form 1.2.3.4:23 or 1.2.3.4_23 into a ip and port
# 
# FIXME: probably could simply be a split on : or _ since those shouldn't 
# appear in the string to be split.
#
sub split_ip_port {
    my($ip_port) = @_;

    if ($ip_port =~ m/(\d+\.\d+\.\d+\.\d+)[\.:_](\d+)/) {
	return($1, $2);
    }
}

#
# read in additional service definitions from the named file
#
sub read_extra {
    my($file) = @_;

    open(FILE, $file) || die "cannot open $file ($!)\n";

    while (<FILE>) {
	chop;
	$_ =~ s/#.*$//;
	next if $_ eq "";

	my($proto, $host, $port, $what) = split(/[ \t]+/, $_);
        $services::port2name{"$proto/$host/$port"} = $what;
    }

    close(FILE);
}

sub get_type {
    my($proto, $ip_a, $port_a, $ip_b, $port_b) = @_;
    my $service_a = '';
    my $service_b = '';
    my $tmp;

    # see whether either port is known
    #
    if (defined($services::port2name{"$proto/*/$port_a"})) {
	$service_a = $services::port2name{"$proto/*/$port_a"};
    }
    if (defined($services::port2name{"$proto/*/$port_b"})) {
	$service_b = $services::port2name{"$proto/*/$port_b"};
    }

    # see whether either port has a special definition that should
    # override the normal known use
    #
    if (defined($services::port2name{"$proto/$ip_a/$port_a"})) {
	$service_a = $services::port2name{"$proto/$ip_a/$port_a"};
    }
    if (defined($services::port2name{"$proto/$ip_b/$port_b"})) {
	$service_b = $services::port2name{"$proto/$ip_b/$port_b"};
    }

    if ($service_a ne '' && $service_b eq '') {
	return(0, $service_a);
    }

    if ($service_a eq '' && $service_b ne '') {
	return(1, $service_b);
    }

    if ($service_a ne '' && $service_b ne '') {
	if ($port_a+0 < $port_b+0) {
	    return(0, $service_a);
	} else {
	    return(0, $service_b);
	}
    }

    return('', '');
}
    
#
# determine service type, service group for a tcp or udp session
# returns (type, group), or (type, '') or ('', '')
#
sub get_tags {
    my($proto, $ip_a, $port_a, $ip_b, $port_b) = @_;

    my $type;
    my $to_server;

    ($to_server, $type) = get_type($proto, $ip_a, $port_a, $ip_b, $port_b);
    return($to_server, '', '') if $type eq '';
    
    return($to_server, $type, '') if ! defined($services::group{$type});

    return($to_server, $type, $services::group{$type});
}

1;
