package x_replay;

require Exporter;
@ISA    = qw(Exporter);
@EXPORT = qw(X_replay_init @X_Request @X_Request_Name);

#use strict;

##############################################################################
#
# Subs to assist in decoding, remapping, and replaying a TCP dump recorded
# X session.
#
# You're not going to have a clue what's going on unless you've got a good
# book on the X protocol to refer to. I'd suggest the "X Protocol Reference
# Manual", volume 0 in O'Reilly's X Window System guide series.
# (ISBN 1-56592-083-X)
#
#
# Copyright 1997, Justin H. Dolske / The Ohio State University
#
##############################################################################

sub X_replay_init {

($new_rid_base, $new_rid_mask, $old_rid_base, $old_rid_mask,
 %server_pixpad, %xserver_pixpad) = @_;

if($old_rid_mask > $new_rid_mask) {
    my $masks;
    $masks = sprintf "%8.8X / %8.8X", $old_rid_mask, $new_rid_mask;
    warn "\nWARNING: The resource mask for the recorded session is larger\n".
         "         the mask assigned by the replay X server. This may\n".
         "         cause large resource IDs, if they exist, to be\n".
         "         incorrectly mapped to new resource IDs.\n".
         "         Old/New mask: " . $masks . "\n";
    }

*l = \$::l;
*s = \$::s;
*new_root_id = \$::new_root_id;
*old_root_id = \$::old_root_id;
*new_root_cmap = \$::new_root_cmap;
*old_root_cmap = \$::old_root_cmap;

*Req_Seq = \$::request_sequence;

$CLIENT = \*::CLIENT;
$SERVER = \*::SERVER;
$XSERVER = \*::XSERVER;
$KEYS = \*::KEYS;

$HaveKeyMap = 0;

&Init_X_subs;
&Init_SymMap;
}


##############################################################################
# These are the subs that handle each request. The n-th sub handles 
# request number n, and the name is the request name with "X_" prepended.
#
sub Init_X_subs {

@X_Request = (
#0 (not X, just to help with bad TCP dumps)
\&FlushZeros,
#1-5
\&X_CreateWindow, \&X_ChangeWindowAttributes, \&X_GetWindowAttributes, undef, undef, 
#6-10
undef, \&X_ReparentWindow, \&X_MapWindow, \&X_MapSubwindows, \&X_UnmapWindow,
#11-15
\&X_UnmapSubwindows, \&X_ConfigureWindow, undef, \&X_GetGeometry, \&X_QueryTree,
#16-20
\&X_InternAtom, \&X_GetAtomName, \&X_ChangeProperty, \&X_DeleteProperty, \&X_GetProperty,
#21-25
undef, \&X_SetSelectionOwner, \&X_GetSelectionOwner, \&X_ConvertSelection, \&X_SendEvent, 
#26-30
undef, undef, \&X_GrabButton, undef, \&X_ChangeActivePointerGrab,
#31-35
undef, undef, undef, undef, undef, 
#36-40
undef, undef, undef, undef, undef,
#41-45
undef, undef, \&X_GetInputFocus, undef, \&X_OpenFont, 
#46-50
\&X_CloseFont, \&X_QueryFont, undef, \&X_ListFonts, undef,
#51-55
undef, undef, \&X_CreatePixmap, \&X_FreePixmap, \&X_CreateGC,
#56-60
\&X_ChangeGC, \&X_CopyGC, undef, undef, \&X_FreeGC,
#61-65
\&X_ClearArea, \&X_CopyArea, \&X_CopyPlane, \&X_PolyPoint, \&X_PolyLine, 
#66-70
\&X_PolySegment, \&X_PolyRectangle, \&X_PolyArc, \&X_FillPoly, \&X_PolyFillRectangle,
#71-75
\&X_PolyFillArc, \&X_PutImage, undef, \&X_PolyText8, undef, 
#76-80
\&X_ImageText8, undef, undef, undef, undef,
#81-85
undef, undef, undef, \&X_AllocColor, \&X_AllocNamedColor, 
#86-90
undef, undef, undef, undef, undef,
#91-95
\&X_QueryColors, undef, \&X_CreateCursor, \&X_CreateGlyphCursor, undef, 
#96-100
\&X_RecolorCursor, undef, \&X_QueryExtension, undef, undef,
#101-105
\&X_GetKeyboardMapping, undef, undef, \&X_Bell, undef, 
#106-110
undef, undef, undef, undef, undef,
#111-115
undef, undef, undef, undef, undef, 
#116-120
undef, undef, undef, \&X_GetModifierMapping, undef
);

@X_Request_Name = qw(
EmptyZero
CreateWindow ChangeWindowAttributes GetWindowAttributes DestroyWindow
DestroySubwindows ChangeSaveSet ReparentWindow MapWindow
MapSubwindows UnmapWindow UnmapSubwindows ConfigureWindow
CirculateWindow GetGeometry QueryTree InternAtom GetAtomName
ChangeProperty DeleteProperty GetProperty ListProperties
SetSelectionOwner GetSelectionOwner ConvertSelection SendEvent
GrabPointer UngrabPointer GrabButton UngrabButton
ChangeActivePointerGrab GrabKeyboard UngrabKeyboard GrabKey UngrabKey
AllowEvents GrabServer UngrabServer QueryPointer GetMotionEvents
TranslateCoords WarpPointer SetInputFocus GetInputFocus QueryKeymap
OpenFont CloseFont QueryFont QueryTextExtents ListFonts
ListFontsWithInfo SetFontPath GetFontPath CreatePixmap FreePixmap
CreateGC ChangeGC CopyGC SetDashes SetClipRectangles FreeGC
ClearArea CopyArea CopyPlane PolyPoint PolyLine PolySegment
PolyRectangle PolyArc FillPoly PolyFillRectangle PolyFillArc
PutImage GetImage PolyText8 PolyText16 ImageText8 ImageText16
CreateColormap FreeColormap CopyColormapAndFree InstallColormap
UninstallColormap ListInstalledColormaps AllocColor AllocNamedColor
AllocColorCells AllocColorPlanes FreeColors StoreColors
StoreNamedColor QueryColors LookupColor CreateCursor
CreateGlyphCursor FreeCursor RecolorCursor QueryBestSize
QueryExtension ListExtensions ChangeKeyboardMapping GetKeyboardMapping
ChangeKeyboardControl GetKeyboardControl Bell ChangePointerControl
GetPointerControl SetScreenSaver GetScreenSaver ChangeHosts
ListHosts SetAccessControl SetCloseDownMode KillClient
RotateProperties ForceScreenSaver SetPointerMapping GetPointerMapping
SetModifierMapping GetModifierMapping xxx_unassigned_120 xxx_unassigned_121
xxx_unassigned_122 xxx_unassigned_123 xxx_unassigned_124 xxx_unassigned_125
xxx_unassigned_126 NoOperation
);

}


#############################################################################
# Loads a translation table for mapping KeySyms to their ASCII values
# 
# Bugs:
#  * not all of the zillions of keysyms are entered here. Extend as
#    needed...
#  * Keypad goofyness...

sub Init_SymMap {

#Load the standard ASCII stuff in set 0
my $i;
for($i=32; $i<127; $i++) {
    $SymMap[$i] = pack "C", $i;
    }

#Load the "Keyboard" set.... See p.350 X Protocol Manual
$SymMap[65280+8]  = "<BS>";
$SymMap[65280+9]  = "<Tab>";
$SymMap[65280+10] = "<LF>";
$SymMap[65280+11] = "<Clear>";
$SymMap[65280+13] = "<Enter>";
$SymMap[65280+19] = "<Pause>";
$SymMap[65280+20] = "<ScrLck>";
$SymMap[65280+27] = "<Esc>";

$SymMap[65280+80] = "<Home>";
$SymMap[65280+81] = "<Left>";
$SymMap[65280+82] = "<Up>";
$SymMap[65280+83] = "<Right>";
$SymMap[65280+84] = "<Down>";
$SymMap[65280+85] = "<PageUp>";
$SymMap[65280+86] = "<PageDown>";
$SymMap[65280+87] = "<End>";
$SymMap[65280+88] = "<Begin>";

$SymMap[65280+97] = "<Print>";
$SymMap[65280+99] = "<Insert>";
$SymMap[65280+106] = "<Help>";
$SymMap[65280+107] = "<Break>";
$SymMap[65280+127] = "<NumLck>";
$SymMap[65280+141] = "<Enter>";

$SymMap[65280+255] = "<Del>";

#do some more in bulk
my @Keyboard = (
		'*', '+', ',', '-', '.', '/', '0', '1', '2', '3', '4', 
		'5', '6', '7', '8', '9', '<?>', '<?>', '<?>', '=', '<F1>', 
		'<F2>', '<F3>', '<F4>', '<F5>', '<F6>', '<F7>', '<F8>', 
		'<F9>', '<F10>', '<F11>', '<F12>', '<F13>', '<F14>', 
		'<F15>', '<F16>', '<F17>', '<F18>', '<F19>', '<F20>', 
		'<F21>', '<F22>', '<F23>', '<F24>', '<F25>', '<F26>', 
		'<F27>', '<F28>', '<F29>', '<F30>', '<F31>', '<F32>', 
		'<F33>', '<F34>', '<F35>', '<Shift>', '<Shift>', 
		'<Control>', '<Control>', '<CapsLck>', '<ShiftLck>', 
		'<Meta>', '<Meta>', '<Alt>', '<Alt>', '<Super>', 
		'<Super>', '<Hyper>', '<Hyper>'
);

$i = 65280 + 170;
my $symbol = shift @Keyboard;
while(defined $symbol) {
    $SymMap[$i] = $symbol;
    $i++;
    $symbol = shift @Keyboard;
    }
}


#############################################################################
# Maps resource identifiers to the new server values. See Remap_Atom for
# dealing with Atoms.
#
sub Remap_ID {

    my $id = shift;

    #added at least for CreateWindow
    if($id < 2) {return $id;}


    if($id == $old_root_id)   {return $new_root_id;}
    if($id == $old_root_cmap) {return $new_root_cmap;}

    #sanity check... If we're remapping an ID, the old ID should have the
    #upper few bits set to the old ID base.
    if(!(($id &~ $old_rid_mask) & $old_rid_base)) {
	print "\nWARNING, this ID has a bad base?\n";
	return $id;
    }

    #just testing an idea...
    #if(($id & 0xFFFF) == 0xFFFF) {return $new_root_id;}

    $id = $id & $old_rid_mask;
    $id = $id | $new_rid_base;
    return $id;
}


############################################################################# 
# Given an Atom (value), it looks the value up in a hash and returns the 
# atom for the new server. Add atoms with Map_Atom.
#
sub Remap_Atom {

my $atom = shift;

if(!defined $atom_map{$atom}) {
    print "[WARNING: No remapping exists!] ";
    return $atom;
    }
return $atom_map{$atom};
}


############################################################################# 
# Add a new mapping between an old atom and it's replay equivilant.
#
sub Map_Atom {

my ($old, $new) = @_;

$atom_map{$old} = $new;

}


#############################################################################
# Process keypresses 
#
# Bugs:
#     * There may be a better way to handle multiple modifiers 
#       (eg Shift-Control-Alt) than this...
#     * Any modifiers other than Shift, ShiftLock, Control, Mod1(Alt) and
#       Mod2 (NumLock) are ignored. This may be a potential problem on
#       keyboards with weird (or different!) modifiers. See p.52 in the
#       aforementioned X Protocol Ref. Manual for the modifiers available
#       in SETofKEYBUTMASK. (there are 13!!)
#     * Doh... Idea: Use the output from GetModifierMapping. This would
#       provide the mapping from Modifier-2 to "NumLock". 

sub XKeyPress {

    my ($keycode, $state) = @_;
    my $key = 0;

    my $oldstate = $state;
    $state = $state & 0x000F; #we will ignore all but the basic modfifiers

    $SymMap[$KeyMap[$keycode][0]] = "<???>" 
	if (!defined $SymMap[$KeyMap[$keycode][0]]);

    #normal...
    $key = $SymMap[$KeyMap[$keycode][0]];
    if (($oldstate & 0x10) && ($keycode >= 79) && ($keycode <= 91)) {
	#numlock is set, and the key is on the number pad
	$key = $SymMap[$KeyMap[$keycode][1]] 
	    unless ($keycode == 86 || $keycode == 82);
	#oops... The "+" and "-" keys don't seem to behave like the others. 
	#we leave them set to the old $key value
    }

    #shift or caps lock...
    if ($state & 0x01 || $state & 0x02) {$key = $SymMap[$KeyMap[$keycode][1]];}

    if ($state > 3) { #found Control and/or Alt
	my $newkey="<";
	$newkey = $newkey . "Ctrl-" if($state & 0x04);
	$newkey = $newkey . "Alt-"  if($state & 0x08);
	$newkey = $newkey . $key . ">";
	$key = $newkey;
    }

    #else {print "WOAH!!\n\n"; $key = $SymMap[$KeyMap[$keycode][0]];} 

    print "        Keypress: \"$key\" (keycode $keycode, modifier mask = $state / $oldstate)\n";

    if ($main::keystroke_file ne '') {
	print $KEYS $key;
    }
}



#############################################################################
#############################################################################



#############################################################################
# 0
sub FlushZeros {

#not really part of the X protocol. tcpdump fills in will nulls whenever
#it has dropped/lost a packet. This is currently a crude attempt to skip
#past them, and hopefully get to good data. This will fail if the data is
#no longer aligned, or if the beginning of a requeust is cutoff.
#
#and, of course, nothing will help if something important was done in the
#missing packet!

my $flush = 0;
read $CLIENT, $flush, 7;

#print "        FLUSHED 8 null bytes?!?!\n";

}


#############################################################################
# 1
# 
# To do:
#  - wouldbe nice to be able to replay 24bpp sessions on 8 bit display
#    (or vice versa?)
sub X_CreateWindow {

my $request = 0;
read $CLIENT, $request, 31;

my($depth,$req_len,$wid,$parent_wid,$x,$y,$width,$height,$border_width,
   $class,$visual,$bitmask) = unpack "C$s$l$l$s$s$s$s$s$s$l$l", $request;

print "        Geometry: $width x $height  $depth bpp  \@($x,$y)\n";
print "        WindowID: " . sprintf "%8.8X", $wid;
$wid = Remap_ID($wid);
print " --> " . sprintf "%8.8X\n", $wid;
print "        ParentID: " . sprintf "%8.8X", $parent_wid;
$parent_wid = Remap_ID($parent_wid);
print " --> " . sprintf "%8.8X\n", $parent_wid;
print "        Visual: " . sprintf "%8.8X\n", $visual;
print "        Bitmask: " . sprintf "%8.8X\n", $bitmask;

if ($main::replay) {
    print $XSERVER pack "CC$s$l$l$s$s$s$s$s$s$l$l", 1,$depth,$req_len,
      $wid,$parent_wid,$x,$y,$width,$height,$border_width,$class,
      $visual,$bitmask;
}

#now we look at the bitmask, if remappable values are to be specified,
#we need to know their position in the list to come.
#(this code also in ChangeWindowAttributes)
my (@remap, $i);
my ($pos, $posmask) = (1,1);
for($i=1; $i<=23; $i++) {
    if($bitmask & $posmask) {
        #set the value to true/false
        $remap[$pos] = $i == 1 || $i == 3 || $i == 14 || $i == 15;
        $pos++;
        }
    $posmask <<= 1;
    }

for($i=1; $i<=$req_len-8; $i++) {
    read $CLIENT, $request, 4;
    if(defined $remap[$i] && $remap[$i] == 1) {
        $request = unpack "$l", $request;
        print "        Remapped: " . sprintf "%8.8X", $request;
        $request = Remap_ID($request);
        print " --> " . sprintf "%8.8X\n", $request;
        print $XSERVER pack "$l", $request
	    if $main::replay;
        }
    else {
        print $XSERVER $request
	    if $main::replay;
        }
    }

}


#############################################################################
# 2
sub X_ChangeWindowAttributes {

my $request = 0;
read $CLIENT, $request, 11;
my($junk,$req_len,$wid,$bitmask) = unpack "C$s$l$l", $request;

print "        WindowID: " . sprintf "%8.8X", $wid;
$wid = Remap_ID($wid);
print " --> " . sprintf "%8.8X\n", $wid;
print "        Bitmask: " . sprintf "%8.8X\n", $bitmask;


print $XSERVER pack "CC$s$l$l", 2,0,$req_len,$wid,$bitmask
    if $main::replay;

#now we look at the bitmask, if remappable values are to be specified,
#we need to know their position in the list to come.
#(this code from CreateWindow)
my (@remap, $i);
my ($pos, $posmask) = (1,1);
for($i=1; $i<=23; $i++) {
    if($bitmask & $posmask) {
        #set the value to true/false
        $remap[$pos] = $i == 1 || $i == 3 || $i == 14 || $i == 15;
        $pos++;
        }
    $posmask <<= 1;
    } 
    
for($i=1; $i<=$req_len-3; $i++) {
    read $CLIENT, $request, 4;
    if(defined $remap[$i] && $remap[$i] == 1) {
        $request = unpack "$l", $request;
        print "        Remapped: " . sprintf "%8.8X", $request;
        $request = Remap_ID($request);
        print " --> " . sprintf "%8.8X\n", $request;
        print $XSERVER pack "$l", $request
	    if $main::replay;
        }
    else {
        print $XSERVER $request
	    if $main::replay;
        }
    } 

}


#############################################################################
# 3
# Bugs:
#   - Ignored
sub X_GetWindowAttributes {

my $request = 0;
read $CLIENT, $request, 7;
my($junk,$req_len,$wid) = unpack "C$s$l", $request;

print "        WindowID: " . sprintf "%8.8X", $wid;
$wid = Remap_ID($wid);
print " --> " . sprintf "%8.8X\n", $wid;

#read $SERVER, $request, 44;
ReadServer(44);

print $XSERVER pack "CC$s", 127,0,1
    if $main::replay;
}


#############################################################################
# 7
sub X_ReparentWindow {

my $request = 0;
read $CLIENT, $request, 15;
my($junk,$req_len,$wid,$parent_wid,$x,$y) = unpack "C$s$l$l$s$s", $request;

print "        WindowID: " . sprintf "%8.8X", $wid;
$wid = Remap_ID($wid);
print " --> " . sprintf "%8.8X\n", $wid;
print "        Parent WindowID: " . sprintf "%8.8X", $parent_wid;
$parent_wid = Remap_ID($parent_wid);
print " --> " . sprintf "%8.8X\n", $parent_wid;
print "        Location: ($x,$y)\n";

print $XSERVER pack "CC$s$l$l$s$s", 7,0,$req_len,$wid,$parent_wid,$x,$y
    if $main::replay;

}


#############################################################################
# 8
sub X_MapWindow {
my $request = 0;
read $CLIENT, $request, 7;
my($junk,$req_len, $wid) = unpack "C$s$l", $request;

print "        WindowID: " . sprintf "%8.8X", $wid;
$wid = Remap_ID($wid);
print " --> " . sprintf "%8.8X\n", $wid;

print $XSERVER pack "CC$s$l", 8,0,$req_len, $wid
    if $main::replay;

#ignore an event...
#ReadXServer(32,19);
}


#############################################################################
# 9
#
# Note: The may generate multiple Notify(?) events from the server, or
# may generate no events is no windows are mapped. Liberal detection and
# ignoring of events when reading from the Xserver takes care of this later.

sub X_MapSubwindows {

my $request = 0;
read $CLIENT, $request, 7;
my($junk,$req_len, $wid) = unpack "C$s$l", $request;

print "        WindowID: " . sprintf "%8.8X", $wid;
$wid = Remap_ID($wid);
print " --> " . sprintf "%8.8X\n", $wid;

print $XSERVER pack "CC$s$l", 9,0,$req_len, $wid
    if $main::replay;

#ignore an event (sigh.. if any, and won't catch 'em all.)
#ReadXServer(32,19);
}


#############################################################################
# 10
sub X_UnmapWindow {

my $request = 0;
read $CLIENT, $request, 7;
my($junk,$req_len, $wid) = unpack "C$s$l", $request;

print "        WindowID: " . sprintf "%8.8X", $wid;
$wid = Remap_ID($wid);
print " --> " . sprintf "%8.8X\n", $wid;

print $XSERVER pack "CC$s$l", 10,0,$req_len, $wid
    if $main::replay;
}


#############################################################################
# 11
sub X_UnmapSubwindows {

my $request = 0;
read $CLIENT, $request, 7;
my($junk,$req_len, $wid) = unpack "C$s$l", $request;

print "        WindowID: " . sprintf "%8.8X", $wid;
$wid = Remap_ID($wid);
print " --> " . sprintf "%8.8X\n", $wid;

print $XSERVER pack "CC$s$l", 11,0,$req_len, $wid
    if $main::replay;
}


#############################################################################
# 12
sub X_ConfigureWindow {

my $request = 0;
read $CLIENT, $request, 11;
my($junk,$req_len,$wid,$bm) = unpack "C$s$l$s", $request;

print "        WindowID: " . sprintf "%8.8X", $wid;
$wid = Remap_ID($wid);
print " --> " . sprintf "%8.8X\n", $wid;
print "        Bitmask: " . sprintf "%8.8X\n", $bm;

print $XSERVER pack "CC$s$l$s$s", 12,0,$req_len,$wid,$bm,0
    if $main::replay;

#now we look at the bitmask, if remappable values are to be specified,
#we need to know their position in the list to come.
my (@remap, $i);
my ($pos, $posmask) = (1,1);
for($i=1; $i<=23; $i++) {
    if($bm & $posmask) {
        #set the value to true/false
        $remap[$pos] = $i == 6;
        $pos++;
        }
    $posmask <<= 1;
    }
    
for($i=1; $i<=$req_len-3; $i++) {
    read $CLIENT, $request, 4;
    if(defined $remap[$i] && $remap[$i] == 1) {
        $request = unpack "$l", $request;
        print "        Remapped: " . sprintf "%8.8X", $request;
        $request = Remap_ID($request);
        print " --> " . sprintf "%8.8X\n", $request;
        print $XSERVER pack "$l", $request
	    if $main::replay;
        }
    else {
        print $XSERVER $request
	    if $main::replay;
        }
    }

#ignore the ConfigureNorify event...
#ReadXServer(32,22);

}


#############################################################################
# 14
# Bugs:
#  - Ignored
sub X_GetGeometry {

my $request = 0;
read $CLIENT, $request, 7;
my($junk,$req_len,$did) = unpack "C$s$l", $request;

print "        Drawable:     " . sprintf "%8.8X --> ", $did;
$did = &Remap_ID($did);
print sprintf "%8.8X\n", $did;

#read $SERVER, $request, 32;
ReadServer(32);

print $XSERVER pack "CC$s", 127,0,1
    if $main::replay;
}


#############################################################################
# 15
sub X_QueryTree {

    my $request = 0;
    read $CLIENT, $request, 7;
    my($junk,$req_len,$wid) = unpack "C$s$l", $request;

    print "        WindowID: " . sprintf "%8.8X", $wid;
    $wid = Remap_ID($wid);
    print " --> " . sprintf "%8.8X\n", $wid;

    my $reply = 0;
    $reply = ReadServer(32);
    my($rep,$junk2,$seq,$rep_len,$rid,$pid,$kids) = unpack "CC$s$l$l$l$s", $reply;
    read $SERVER, $reply, 4*$rep_len;

    if ($main::replay) {
	print $XSERVER pack "CC$s$l", 15,0,2,$wid;
	my $reply2 = 0;
	$reply2 = ReadXServer(32);

	my($rep2,$junk3,$seq2,$rep_len2,$rid2,$pid2,$kids2) = 
	    unpack "CC$s$l$l$l$s", $reply2;
	read $XSERVER, $reply2, 4*$rep_len2;

	print "        Root-of:   " . sprintf "%8.8X", $rid;
	print sprintf " --> %8.8X\n", $rid2;
	print "        Parent-of: " . sprintf "%8.8X", $pid; 
	print sprintf " --> %8.8X\n", $pid2;
	print "        Kids: $rep_len --> $rep_len2\n";
	print "        Kids: $kids --> $kids2\n";
    }
}


#############################################################################
# 16

sub X_InternAtom {
    my($atom, $atom2, $ifexist, $junk, $junk2, $name, $name_len, $rep, 
       $rep2, $rep_len, $rep_len2, $reply, $req_len, $request, $seq, $seq2);

    ($request, $name) = (0,0);
    read $CLIENT, $request, 7;
    ($ifexist, $req_len, $name_len, $junk) = unpack "C$s$s$s", $request;
    read $CLIENT, $name, $name_len;
    $junk = "";
    read $CLIENT, $junk, 4*($req_len-2)-$name_len;
    print "        IfExists: $ifexist\n";
    print "        Name: $name\n";

    #see what the recorded server said
    $reply = 0;
    #read $SERVER, $reply, 32;
    $reply = ReadServer(32);
    ($rep2,$junk2,$seq2,$rep_len2,$atom2) = unpack "CC$s$l$l", $reply;

    if ($main::replay) {
	print $XSERVER pack "CC$s$s$s", 16,$ifexist, $req_len, $name_len, 0;
	print $XSERVER "$name$junk";

	$reply = ReadXServer(32);
	($rep,$junk,$seq,$rep_len,$atom) = unpack "CC$s$l$l", $reply;
    } else {
	$atom = $atom2;
    }

    print "        Reply atoms: " . sprintf "%8.8X ", $atom2;
    print sprintf "/ %8.8X\n", $atom;

    Map_Atom($atom2, $atom);
    if(($atom == 0 || $atom2 == 0) && ($atom != $atom2)) {
	print "WARNING: Atom does not exist in ";
	if($atom == 0) {print "Xserver.\n";}
	else {          print "recorded server.\n";}
    }
}


#############################################################################
# 17
#
# Ignored for the moment...
sub X_GetAtomName {

    my($atom, $junk, $name, $name_len, $rep, $rep_len, $reply, $req_len, 
       $request, $seq);

    read $CLIENT, $request, 7;
    ($junk, $req_len, $atom) = unpack "C$s$l", $request;
    
    print "        Atom: " . sprintf "%8.8X ", $atom;
    $atom = Remap_Atom($atom);
    print "--> " . sprintf "%8.8X\n", $atom;
    
    $reply = ReadServer(32);
    ($rep,$junk,$seq,$rep_len,$name_len) = unpack "CC$s$l$s", $reply;

    read $SERVER, $name, $name_len;
    print "        OldName: $name\n";
    read $SERVER, $junk, 4*$req_len - $name_len;

    print $XSERVER pack "CC$s", 127,0,1
	if $main::replay;
}


#############################################################################
# 18
#
# Bugs:
#   - seems to be broken on at least 1 request, hacked to temp fix
#   - hmm. at some point i had the prop/type to NOT be remapped. i don't
#     know why i did this. :) perhaps because i was (wrongly) using remap_id 
#     perhaps because remapping atoms was unreliable at time? *Sigh*

sub X_ChangeProperty {

my ($request,$data) = (0,0);
read $CLIENT, $request, 23;
my($mode,$req_len,$wid,$property,$type,$format,$j1,$j2,$j3,$data_len) =
    unpack "C$s$l$l$l"."CCCC$l", $request;
read $CLIENT, $data, 4*($req_len-6);

print "        WindowID: " . sprintf "%8.8X", $wid;
$wid = Remap_ID($wid);
print " --> " . sprintf "%8.8X\n", $wid;
#print "        Property/type: " . sprintf "%8.8X / %8.8X\n",$property,$type;
print "        Property: " . sprintf "%8.8X", $property;
$property = Remap_Atom($property);
print " --> " . sprintf "%8.8X\n", $property;
print "        Type: " . sprintf "%8.8X", $type;
$type = Remap_Atom($type);
print " --> " . sprintf "%8.8X\n", $type;

###HACK###
# 1 is for the 24 bit test sessions, 2 for the others (16 bpp)
if($property == 151 || $property == 133) {print "SKIPPED! (**1**)\n"; return;}
if($property == 207 || $property == 208 || $property == 180) {print "SKIPPED!(**2**)\n"; return;}

    if ($main::replay) {
	print $XSERVER pack "CC$s$l$l$l"."CCCC$l", 18,$mode,$req_len,$wid,
            $property,$type,$format,0,0,0,$data_len;
	print $XSERVER $data;
    }
}


#############################################################################
# 19
# 
# Ignored for now...
sub X_DeleteProperty {

    my($atom, $junk, $len, $request, $wid);

    $request = 0;
    ($junk,$len,$wid,$atom) = (0,0,0,0);

    read $CLIENT, $request, 11;
    ($junk,$len,$wid,$atom) = unpack "C$s$l$l", $request;
    print "        WindowID: " . sprintf "%8.8X\n", $wid;
    print "        Atom: " . sprintf "%8.8X\n", $atom;
    print "        (IGNORED)\n";
    
    print $XSERVER pack "CC$s", 127,0,1
	if $main::replay;
}


#############################################################################
# 20
sub X_GetProperty {
    my($B_after, $B_after2, $delete, $fmt, $fmt2, $long_len, $long_offset, 
       $property, $rep, $rep2, $rep_len, $rep_len2, $reply, $req_len, 
       $request, $seq, $seq2, $type, $type2, $val_len, $val_len2, $win_id);

    read $CLIENT, $request, 23;  #23 + 4 mystery bytes!
    ($delete,$req_len,$win_id,$property,$type,$long_offset,$long_len) =
	unpack "C$s$l$l$l$l$l", $request;
    print "        Delete: $delete\n";
    print "        Win_ID: " . sprintf "%8.8X ", $win_id;
    $win_id = Remap_ID($win_id);
    print "--> " . sprintf "%8.8X\n", $win_id;
    print "        Property: " . sprintf "%8.8X ", $property;
    $property = Remap_Atom($property);
    print "--> " . sprintf "%8.8X\n", $property;
    print "        Type    : " . sprintf "%8.8X\n", $type;

    #see what the recorded server said
    $reply = 0;
    $reply = ReadServer(32);
    ($rep2,$fmt2,$seq2,$rep_len2,$type2,$B_after2,$val_len2) =
	unpack "CC$s$l$l$l$l", $reply;

    if ($main::replay) {
	print $XSERVER pack "CC$s$l$l$l$l$l",20,$delete,$req_len,$win_id,
  	    $property, $type,$long_offset,$long_len;

        #see what the server says

	$reply = ReadXServer(32);
	($rep,$fmt,$seq,$rep_len,$type,$B_after,$val_len) =
	    unpack "CC$s$l$l$l$l", $reply;
	die "Bad Xserver response.\n" if($rep != 1);
    } else {
	$type = $type2;
	$fmt = $fmt2;
	$rep_len = $rep_len2;
    }

    print "        Reply Atoms (Xser/ser): " . sprintf "%8.8X", $type;
    print sprintf " / %8.8X\n", $type2;

    die "Bad recorded server response.\n" if($rep2 != 1);
    warn "Format mismatch! ($fmt2 / $fmt)\n" if($fmt != $fmt2);
    warn "Length mismatch! ($rep_len2 / $rep_len)\n" if($rep_len != $rep_len2);

    print "        Reply Lengths (Xserver/server): $rep_len / $rep_len2\n";

    #ignore this part of the reply
    read $XSERVER, $reply, 4*$rep_len
	if $main::replay;  

    #ignore this part of the reply
    read $SERVER, $reply, 4*$rep_len2;
}


#############################################################################
# 22
sub X_SetSelectionOwner {

my $request = 0;
read $CLIENT, $request, 15;
my ($junk,$req_len,$wid,$atom,$timestamp) = unpack "C$s$l$l$l", $request;

print "        WindowID: " . sprintf "%8.8X", $wid;
$wid = Remap_ID($wid);
print " --> " . sprintf "%8.8X\n", $wid;
print "        Selection: " . sprintf "%8.8X ", $atom;
$atom = Remap_Atom($atom);
print "--> " . sprintf "%8.8X\n", $atom;

print $XSERVER pack "CC$s$l$l$l", 22,0,$req_len,$wid,$atom,$timestamp
    if $main::replay;

}


#############################################################################
# 23
sub X_GetSelectionOwner {

    my $request = 0;
    read $CLIENT, $request, 7;
    my($junk,$req_len,$atom) = unpack "C$s$l", $request;

    print "        Selection: " . sprintf "%8.8X ", $atom;
    $atom = Remap_Atom($atom);
    print "--> " . sprintf "%8.8X\n", $atom;

    if ($main::replay) {
	print $XSERVER pack "CC$s$l", 23,0,$req_len,$atom;

        $reply = 0;
        $reply = ReadXServer(32);
        ($op,$j,$seq,$len,$wid) = unpack "CC$s$l$l", $reply;
        print "        Xserver Owner:  " . sprintf "%8.8X\n", $wid;
    }

    my $reply = ReadServer(32);
    my ($op,$j,$seq,$len,$wid) = unpack "CC$s$l$l", $reply;
    print "        Recorded Owner: " . sprintf "%8.8X\n", $wid;

}


#############################################################################
# 24
#    Ignored
sub X_ConvertSelection {

my $request = 0;
read $CLIENT, $request, 23;

print $XSERVER pack "CC$s", 127,0,1
    if $main::replay;
}


#############################################################################
# 25
# Bugs: The state of the pointer location may affect the server's action?
sub X_SendEvent {

    my $request = 0;
    read $CLIENT, $request,11;
    my($propigate,$req_len,$wid,$mask) = unpack "C$s$l$l", $request;
    read $CLIENT, $request, 32;

    print "        WindowID: " . sprintf "%8.8X", $wid;
    $wid = Remap_ID($wid);
    print " --> " . sprintf "%8.8X\n", $wid;

    if ($main::replay) {
	print $XSERVER pack "CC$s$l$l", 25,$propigate,$req_len,$wid,$mask;
	print $XSERVER $request;
    }

}


#############################################################################
# 28
#  - Ignored
sub X_GrabButton {

    my $request = 0;
    read $CLIENT, $request, 23;

    #this request is ignored for replay

    print $XSERVER pack "CC$s", 127,0,1
	if $main::replay;
}


#############################################################################
# 30
sub X_ChangeActivePointerGrab {

    my $request = 0;
    read $CLIENT, $request, 15;
    my($junk,$req_len,$cid,$timestamp,$eventmask) = 
	unpack "C$s$l$l$s", $request;

    print "        CursorID:     " . sprintf "%8.8X --> ", $cid;
    $cid = &Remap_ID($cid);
    print sprintf "%8.8X\n", $cid;

    print $XSERVER pack "CC$s$l$l$s$s", 30,0,$req_len,$cid,
    	$timestamp,$eventmask,0
	if $main::replay;
}


#############################################################################
# 43
# This request is ignored
sub X_GetInputFocus {

my $request = 0;
read $CLIENT, $request, 3;
print "        IGNORED.\n";

ReadServer(32);

print $XSERVER pack "CC$s", 127,0,1
    if $main::replay;
}


#############################################################################
# 45
#
# Bugs: Mysterious replies(events)
#       ^--fixed. "cursor" opens seemed to have a 32 byte request. Looks
#          like this was just a weird interaction w/replies to unsuppored
#          requests (eg, extensions)
sub X_OpenFont {

    my ($request, $name) = (0,0);
    read $CLIENT, $request, 11;
    my($junk, $req_len, $fid, $name_len) = unpack "C$s$l$s", $request;
    read $CLIENT, $name, $name_len;
    $junk = "";
    read $CLIENT, $junk, 4*($req_len-3)-$name_len;
    print "        Name: $name\n";
    print "        FontID: " . sprintf "%8.8X --> ", $fid;
    $fid = &Remap_ID($fid);
    print sprintf "%8.8X\n", $fid;

    if ($main::replay) {
	print $XSERVER pack "CC$s$l$s$s", 45,0,$req_len,$fid,$name_len,0;
	print $XSERVER "$name$junk";
    }
}


#############################################################################
# 46
sub X_CloseFont {

    my $request = 0;
    read $CLIENT, $request, 7;
    my($junk, $req_len, $fid) = unpack "C$s$l", $request;

    print "        FontID: " . sprintf "%8.8X --> ", $fid;
    $fid = &Remap_ID($fid);
    print sprintf "%8.8X\n", $fid;

    print $XSERVER pack "CC$s$l", 46,0,$req_len, $fid
	if $main::replay;

}


#############################################################################
# 47
#
# Bugs:
#   - There's an ATOM in the reply, so we should really process it 
#     instead of ignoring it. But hey, I'm a lazy Perl programmer.
#   - Error in documentation? How is req_len only *7* + extensions? 
#     the actual size looks more like *15* + extensions!
sub X_QueryFont {

    my($fid, $junk, $rep, $rep_len, $rep_len2, $reply, $req_len, 
       $request, $seq);

    read $CLIENT, $request, 7;
    ($junk, $req_len, $fid) = unpack "C$s$l", $request;

    print "        FontID: " . sprintf "%8.8X --> ", $fid;
    $fid = &Remap_ID($fid);
    print sprintf "%8.8X\n", $fid;

    $reply = ReadServer(8);
    ($rep,$junk,$seq,$rep_len) = unpack "CC$s$l", $reply;

    read $SERVER, $reply, 4*($rep_len-2);
    read $SERVER, $reply, 32; ###HACK###?

    if ($main::replay) {
	print $XSERVER pack "CC$s$l", 47,0,$req_len, $fid;

	$reply = ReadXServer(8);
	($rep,$junk,$seq,$rep_len2) = unpack "CC$s$l", $reply; 

	read $XSERVER, $reply, 4*($rep_len2-2);
	read $XSERVER, $reply, 32;
    } else {
	$rep_len2 = $rep_len;
    }

    if($rep_len != $rep_len2) {
	print "WARNING: Reply lengths differed... $rep_len --> $rep_len2\n";
    } 
}


#############################################################################
# 49
# 
# This request is ignored
sub X_ListFonts {

my $request = 0;
read $CLIENT, $request, 7;
my($junk, $req_len, $max_names, $pat_len) = unpack "C$s$s$s", $request;
read $CLIENT, $request, $pat_len;
print "        Pattern: $request\n";
read $CLIENT, $request, 4*($req_len-2)-$pat_len;

my $reply = 0;
$reply = ReadServer(32);
my($type,$j,$seq,$rep_len,$num_str) = unpack "CC$s$l$l", $reply;
print "        Fonts found: $num_str\n";
read $SERVER, $reply, $rep_len*4;

print $XSERVER pack "CC$s", 127,0,1
    if $main::replay;
}


#############################################################################
# 53
sub X_CreatePixmap {

my $request = 0;
read $CLIENT, $request, 15;
my($depth, $req_len, $pid, $drawable, $width, $height) =
    unpack "C$s$l$l$s$s", $request;
print "        Type: $width x $height x $depth bpp\n";
print "        PixmapID: " . sprintf "%8.8X --> ", $pid;
$pid = &Remap_ID($pid);
print sprintf "%8.8X\n", $pid;
print "        Drawable: " . sprintf "%8.8X --> ", $drawable;
$drawable = &Remap_ID($drawable);
print sprintf "%8.8X\n", $drawable;

print $XSERVER pack "CC$s$l$l$s$s", 53, $depth,$req_len, 
               $pid, $drawable, $width, $height
		   if $main::replay;
}


#############################################################################
# 54
sub X_FreePixmap {

my $request = 0;
read $CLIENT, $request, 7;
my($junk, $req_len, $pid) = unpack "C$s$l", $request;

print "        PixmapID: " . sprintf "%8.8X --> ", $pid;
$pid = &Remap_ID($pid);
print sprintf "%8.8X\n", $pid;

print $XSERVER pack "CC$s$l", 54,0,$req_len, $pid
    if $main::replay;
}


#############################################################################
# 55
# Note: This code is replicated in the ChangeGC sub.
sub X_CreateGC {

my ($junk,$request,$req_len,$cid,$drawable,$req_tmp,$bm) = 
    (0,0,0,0,0,0,0);

read $CLIENT, $request, 15;
($junk,$req_len,$cid,$drawable,$bm) = unpack "C$s$l$l$l", $request;

print "        CID:      " . sprintf "%8.8X", $cid;
$cid = Remap_ID($cid);
print " --> " . sprintf "%8.8X\n", $cid;
print "        Drawable: " . sprintf "%8.8X", $drawable;
$drawable = Remap_ID($drawable);
print " --> " . sprintf "%8.8X\n", $drawable;
print "        Mask: " . sprintf "%8.8X\n", $bm;

#now we look at the bitmask, if remappable values are to be specified,
#we need to know their position in the list to come.
my @remap;
my ($pos, $mask) = (1,1);
for($i=1; $i<=23; $i++) {
    if($bm & $mask) {
        #set the value to true/false
        $remap[$pos] = $i == 11 || $i == 12 || $i == 15 || $i == 20;
        $pos++;
        }
    $mask <<= 1;
    }

print $XSERVER pack "CC$s$l$l$l", 55, 0, $req_len, $cid, $drawable, $bm
    if $main::replay;

for($i=1; $i<= $req_len-4; $i++) {
    read $CLIENT, $request, 4;
    if(defined $remap[$i] && $remap[$i] == 1) {
        $request = unpack "$l", $request;
        print "        Remapped: " . sprintf "%8.8X", $request;
        $request = Remap_ID($request);
        print " --> " . sprintf "%8.8X\n", $request;
        print $XSERVER pack "$l", $request
	    if $main::replay;
        }
    else {
        print $XSERVER $request
	    if $main::replay;
        }
    }
}


#############################################################################
# 56
# Note: This code is based on the CreateGC sub.
sub X_ChangeGC {

my ($junk,$request,$req_len,$cid,$req_tmp,$bm) =
    (0,0,0,0,0,0);

read $CLIENT, $request, 11;
($junk,$req_len,$cid,$bm) = unpack "C$s$l$l", $request;

print "        CID:      " . sprintf "%8.8X", $cid;
$cid = Remap_ID($cid);
print " --> " . sprintf "%8.8X\n", $cid;
print "        Mask: " . sprintf "%8.8X\n", $bm;

#now we look at the bitmask, if remappable values are to be specified,
#we need to know their position in the list to come.
my @remap;
my ($pos, $mask) = (1,1);
for($i=1; $i<=23; $i++) {
    if($bm & $mask) {
        #set the value to true/false
        $remap[$pos] = $i == 11 || $i == 12 || $i == 15 || $i == 20;
        $pos++;
        }
    $mask <<= 1;
    }

print $XSERVER pack "CC$s$l$l", 56, 0, $req_len, $cid, $bm
    if $main::replay;

for($i=1; $i<= $req_len-3; $i++) {
    read $CLIENT, $request, 4;
    if(defined $remap[$i] && $remap[$i] == 1) {
        $request = unpack "$l", $request;
        print "        Remapped: " . sprintf "%8.8X", $request;
        $request = Remap_ID($request);
        print " --> " . sprintf "%8.8X\n", $request;
        print $XSERVER pack "$l", $request
	    if $main::replay;
        }
    else {
        print $XSERVER $request
	    if $main::replay;
        }
    }

}


#############################################################################
# 57
sub X_CopyGC {
      
my ($junk,$request,$req_len,$src_cid,$dst_cid,$req_tmp,$bm) =
    (0,0,0,0,0,0);
      
read $CLIENT, $request, 15;
($junk,$req_len,$src_cid,$dst_cid,$bm) = unpack "C$s$l$l$l", $request;
      
print "        Src-CID:      " . sprintf "%8.8X", $src_cid;
$src_cid = Remap_ID($src_cid);
print " --> " . sprintf "%8.8X\n", $src_cid;
print "        Dst-CID:      " . sprintf "%8.8X", $dst_cid;
$dst_cid = Remap_ID($dst_cid);   
print " --> " . sprintf "%8.8X\n", $dst_cid;
print "        Mask: " . sprintf "%8.8X\n", $bm;
      
print $XSERVER pack "CC$s$l$l$l", 57, 0, $req_len, $src_cid, $dst_cid, $bm
    if $main::replay;
      
}

#############################################################################
# 60
sub X_FreeGC {

my $request = 0;
read $CLIENT, $request, 7;
my($junk, $req_len, $gid) = unpack "C$s$l", $request;
print "        GContext:     " . sprintf "%8.8X --> ", $gid;
$gid = &Remap_ID($gid);
print sprintf "%8.8X\n", $gid;

print $XSERVER pack "CC$s$l", 60, 0, $req_len, $gid
    if $main::replay;
}


#############################################################################
# 61
sub X_ClearArea {

my $request = 0;
read $CLIENT, $request, 15;
my($exp,$req_len,$wid,$x,$y,$width,$height)=unpack "C$s$l$s$s$s$s", $request;

print "        WindowID: " . sprintf "%8.8X", $wid;
$wid = Remap_ID($wid);
print " --> " . sprintf "%8.8X\n", $wid;
print "        Clearing $width x $height block \@ ($x,$y)\n";

print $XSERVER pack "CC$s$l$s$s$s$s", 61,$exp,4,$wid,$x,$y,$width,$height
    if $main::replay;

#ignore the NoExpose event...
#ReadXServer(32,14);
}


#############################################################################
# 62
sub X_CopyArea {

my $request = 0;
read $CLIENT, $request, 27;
my($junk,$req_len,$src_did,$dst_did,$gid,$srcx,$srcy,$dstx,$dsty,
   $width,$height) = unpack "C$s$l$l$l$s$s$s$s$s$s", $request;

print "        Src Drawable:     " . sprintf "%8.8X --> ", $src_did;
$src_did = &Remap_ID($src_did);
print sprintf "%8.8X\n", $src_did;
print "        Dst Drawable:     " . sprintf "%8.8X --> ", $dst_did;
$dst_did = &Remap_ID($dst_did);
print sprintf "%8.8X\n", $dst_did;
print "        GContext:     " . sprintf "%8.8X --> ", $gid;
$gid = &Remap_ID($gid);
print sprintf "%8.8X\n", $gid;
print "        Copying $width x $height block \@ ($srcx,$srcy) to ($dstx,$dsty)\n";

print $XSERVER pack "CC$s$l$l$l$s$s$s$s$s$s", 62,0,7,$src_did,$dst_did,$gid,
               $srcx,$srcy,$dstx,$dsty,$width,$height
		   if $main::replay;
}


#############################################################################
# 63
sub X_CopyPlane {

my $request = 0;
read $CLIENT, $request, 31;
my($junk,$req_len,$src_did,$dst_did,$gid,$srcx,$srcy,$dstx,$dsty,
   $width,$height,$bitplane) = unpack "C$s$l$l$l$s$s$s$s$s$s$l", $request;

print "        Src Drawable:     " . sprintf "%8.8X --> ", $src_did;
$src_did = &Remap_ID($src_did);
print sprintf "%8.8X\n", $src_did;
print "        Dst Drawable:     " . sprintf "%8.8X --> ", $dst_did;
$dst_did = &Remap_ID($dst_did);
print sprintf "%8.8X\n", $dst_did;
print "        GContext:     " . sprintf "%8.8X --> ", $gid;
$gid = &Remap_ID($gid);
print sprintf "%8.8X\n", $gid;
print "        Copying $width x $height block \@ ($srcx,$srcy), plane $bitplane to ($dstx,$dsty)\
n";

print $XSERVER pack "CC$s$l$l$l$s$s$s$s$s$s$l", 63,0,8,$src_did,$dst_did,$gid,
               $srcx,$srcy,$dstx,$dsty,$width,$height,$bitplane
		   if $main::replay;

}


#############################################################################
# 64
# NOTE: THIS REQUEST IS UNTESTED (based on X_PolyLine code)
# It should certainly work though.
sub X_PolyPoint {

print "        ***UNTESTED CODE***\n";

my ($request, $points) = (0,0);
read $CLIENT, $request, 11;
my($coord, $req_len, $did, $gid) = unpack "C$s$l$l", $request;
read $CLIENT, $points, 4*($req_len-3);

print "        Drawable:     " . sprintf "%8.8X --> ", $did;
$did = &Remap_ID($did);
print sprintf "%8.8X\n", $did;
print "        GContext:     " . sprintf "%8.8X --> ", $gid;
$gid = &Remap_ID($gid);
print sprintf "%8.8X\n", $gid;

print $XSERVER pack "CC$s$l$l", 64,$coord,$req_len,$did,$gid
    if $main::replay;
print $XSERVER $points
    if $main::replay;

}


#############################################################################
# 65
sub X_PolyLine {

my ($request, $points) = (0,0);
read $CLIENT, $request, 11;
my($coord, $req_len, $did, $gid) = unpack "C$s$l$l", $request;
read $CLIENT, $points, 4*($req_len-3);

print "        Drawable:     " . sprintf "%8.8X --> ", $did;
$did = &Remap_ID($did);
print sprintf "%8.8X\n", $did;
print "        GContext:     " . sprintf "%8.8X --> ", $gid;
$gid = &Remap_ID($gid);
print sprintf "%8.8X\n", $gid;

print $XSERVER pack "CC$s$l$l", 65,$coord,$req_len,$did,$gid
    if $main::replay;
print $XSERVER $points
    if $main::replay;

}


#############################################################################
# 66
sub X_PolySegment {

my $request = 0;
read $CLIENT, $request, 11;
my($junk, $req_len, $did, $gid) = unpack "C$s$l$l", $request;
    
print "        Drawable:     " . sprintf "%8.8X --> ", $did;
$did = &Remap_ID($did);
print sprintf "%8.8X\n", $did;
print "        GContext:     " . sprintf "%8.8X --> ", $gid;
$gid = &Remap_ID($gid);
print sprintf "%8.8X\n", $gid;
    
print $XSERVER pack "CC$s$l$l", 66,0,$req_len,$did,$gid
    if $main::replay;

read $CLIENT, $request, 4*($req_len-3);
print $XSERVER $request
    if $main::replay;
}


#############################################################################
# 67
sub X_PolyRectangle {

my $request = 0;
read $CLIENT, $request, 11;
my($junk, $req_len, $did, $gid) = unpack "C$s$l$l", $request;

print "        Drawable:     " . sprintf "%8.8X --> ", $did;
$did = &Remap_ID($did);
print sprintf "%8.8X\n", $did;
print "        GContext:     " . sprintf "%8.8X --> ", $gid;
$gid = &Remap_ID($gid);
print sprintf "%8.8X\n", $gid;

print $XSERVER pack "CC$s$l$l", 67,0,$req_len,$did,$gid
    if $main::replay;

read $CLIENT, $request, 4*($req_len-3);
print $XSERVER $request
    if $main::replay;
}


#############################################################################
# 68
# NOTE: THIS REQUEST IS UNTESTED (based loosely on X_PolyLine code)
# It should certainly work though.
sub X_PolyArc {

print "        ***UNTESTED CODE***\n";

my ($request, $arcs) = (0,0);
read $CLIENT, $request, 11;
my($dummy, $req_len, $did, $gid) = unpack "C$s$l$l", $request;
read $CLIENT, $arcs, 4*($req_len-3);

print "        Drawable:     " . sprintf "%8.8X --> ", $did;
$did = &Remap_ID($did);
print sprintf "%8.8X\n", $did;
print "        GContext:     " . sprintf "%8.8X --> ", $gid;
$gid = &Remap_ID($gid);
print sprintf "%8.8X\n", $gid;

print $XSERVER pack "CC$s$l$l", 68,0,$req_len,$did,$gid
    if $main::replay;
print $XSERVER $arcs
    if $main::replay;

}


#############################################################################
# 69
# NOTE: THIS REQUEST IS UNTESTED (based loosely on X_PolyLine code)
# It should work though.
sub X_FillPoly {

print "        ***UNTESTED CODE***\n";

my ($request, $points) = (0,0);
read $CLIENT, $request, 15;
my($dummy, $req_len, $did, $gid,$shape,$coord) = 
    unpack "C$s$l$l"."CC", $request;
read $CLIENT, $points, 4*($req_len-4);

print "        Drawable:     " . sprintf "%8.8X --> ", $did;
$did = &Remap_ID($did);
print sprintf "%8.8X\n", $did;
print "        GContext:     " . sprintf "%8.8X --> ", $gid;
$gid = &Remap_ID($gid);
print sprintf "%8.8X\n", $gid;

print $XSERVER pack "CC$s$l$l"."CC$s", 69,0,$req_len,$did,$gid,$shape,$coord,0
    if $main::replay;
print $XSERVER $points
    if $main::replay;

}


#############################################################################
# 70
sub X_PolyFillRectangle {

my $request = 0;
read $CLIENT, $request, 11;
my($junk, $req_len, $did, $gid) = unpack "C$s$l$l", $request;

print "        Drawable:     " . sprintf "%8.8X --> ", $did;
$did = &Remap_ID($did);
print sprintf "%8.8X\n", $did;
print "        GContext:     " . sprintf "%8.8X --> ", $gid;
$gid = &Remap_ID($gid);
print sprintf "%8.8X\n", $gid;

print $XSERVER pack "CC$s$l$l", 70,0,$req_len,$did,$gid
    if $main::replay;

read $CLIENT, $request, 4*($req_len-3);
print $XSERVER $request
    if $main::replay;
}


#############################################################################
# 71 
# NOTE: THIS REQUEST IS UNTESTED (based on X_PolyArc)
# It should certainly work though.
sub X_PolyFillArc { 
 
print "        ***UNTESTED CODE***\n";
 
my ($request, $arcs) = (0,0);
read $CLIENT, $request, 11;
my($dummy, $req_len, $did, $gid) = unpack "C$s$l$l", $request;
read $CLIENT, $arcs, 4*($req_len-3);

print "        Drawable:     " . sprintf "%8.8X --> ", $did;
$did = &Remap_ID($did);
print sprintf "%8.8X\n", $did;
print "        GContext:     " . sprintf "%8.8X --> ", $gid;
$gid = &Remap_ID($gid);
print sprintf "%8.8X\n", $gid;

print $XSERVER pack "CC$s$l$l", 71,0,$req_len,$did,$gid
    if $main::replay;
print $XSERVER $arcs
    if $main::replay;
 
}


#############################################################################
# 72
#
# Bugs:
#  - Hardcoded pad of 32 if not found
#  - Generally messy :-(
sub X_PutImage {

my ($request,$data) = (0,0);
read $CLIENT, $request, 23;
my($format, $req_len, $did, $gid, $width, $height, $dst_x, $dst_y, $lpad,
   $depth) = unpack "C$s$l$l$s$s$s$s"."CC", $request;
read $CLIENT, $data, 4*($req_len-6);

print "        Format: $format\n";
print "        Info: $width x $height  $depth bpp \@ ($dst_x,$dst_y)\n";
print "        Left pad bits: $lpad\n";
print "        Data size: ", 4*($req_len-6), "\n";
print "        Drawable:     " . sprintf "%8.8X --> ", $did;
$did = &Remap_ID($did);
print sprintf "%8.8X\n", $did;
print "        GContext:     " . sprintf "%8.8X --> ", $gid;
$gid = &Remap_ID($gid);
print sprintf "%8.8X\n", $gid;

if(!defined $xserver_pixpad{$depth}) {
    print "NOTICE: Xserver's right-pad unknown, guessing 32.\n";
    $xserver_pixpad{$depth} = 32;
    }
if(!defined $server_pixpad{$depth}) {
    print "NOTICE: Server's right-pad unknown, guessing 32.\n";
    $server_pixpad{$depth} = 32;
    }

print "        Server pad:  $server_pixpad{$depth}\n";
print "        Xserver pad: $xserver_pixpad{$depth}\n";

#recalculate req_len, and force lpad to 0.
#first, find length of each row
$req_len = $width + $xserver_pixpad{$depth}-
    ($width % $xserver_pixpad{$depth})%$xserver_pixpad{$depth};
$req_len *= $height;
$req_len /= 8; #change to bytes
$req_len = ($req_len + (4-($req_len % 4))%4)/4 + 6;

print "    New data size: ", 4*($req_len-6), "\n";
print $XSERVER pack "CC$s$l$l$s$s$s$s"."CC$s", 72, $format, $req_len,
      $did, $gid, $width, $height, $dst_x, $dst_y, 0, $depth, 0
	  if $main::replay;


#send the pixels
if($format == 0 || $format == 1) {  ###HACK### allows 1bpp, format 1
    #how many bits per row (bpr)?
    $w = $width; $bpr = 0;
    while($w >= 0) {
        $w -= $server_pixpad{$depth};
        $bpr += $server_pixpad{$depth};
        }
    #make data a string of 1/0
    $data = unpack "b*", $data;

    for($i=0; $i<$height; $i++) {
        $row = substr $data, $bpr*$i+$lpad, $width;
        $row = pack "b$xserver_pixpad{$depth}", $row;
        print unpack "b*", $row;
        print "\n";
        print $XSERVER $row
	    if $main::replay;
        }
    }

}


#############################################################################
# 74
# Might be nice to extract and show (textual) what the string is.
sub X_PolyText8 {

my $request = 0;
read $CLIENT, $request, 15;
my($junk,$req_len,$did,$gid,$x,$y) = unpack "C$s$l$l$s$s", $request;
read $CLIENT, $request, 4*($req_len-4); #strings'n'stuff

print "        Drawable:     " . sprintf "%8.8X --> ", $did;
$did = &Remap_ID($did);
print sprintf "%8.8X\n", $did;
print "        GContext:     " . sprintf "%8.8X --> ", $gid;
$gid = &Remap_ID($gid);
print sprintf "%8.8X\n", $gid;

print $XSERVER pack "CC$s$l$l$s$s", 74,0,$req_len,$did,$gid,$x,$y
    if $main::replay;
print $XSERVER $request
    if $main::replay;
}


#############################################################################
# 76
sub X_ImageText8 {

    my($request,$string) = (0,0);
    read $CLIENT, $request, 15;
    my($str_len, $req_len, $did,$gid,$x,$y) = unpack "C$s$l$l$s$s", $request;
    read $CLIENT, $string, 4*($req_len-4);

    print "        Drawable:     " . sprintf "%8.8X --> ", $did;
    $did = &Remap_ID($did);
    print sprintf "%8.8X\n", $did;
    print "        GContext:     " . sprintf "%8.8X --> ", $gid;
    $gid = &Remap_ID($gid);
    print sprintf "%8.8X\n", $gid;

    if ($main::replay) {
	print $XSERVER pack "CC$s$l$l$s$s", 76,$str_len, $req_len, $did,$gid,$x,$y;
	print $XSERVER $string;
    }

    $string = substr $string,0,$str_len;
    print "        String: $string\.\n";

    if($main::replay && length $string > 0 && length $string < 2 && $string ne " ") {
	select undef,undef,undef,.1;
    }
}


#############################################################################
# 84
#
# - Ignored

sub X_AllocColor {

my ($request,$cname) = (0,0);
read $CLIENT, $request, 15;
my($junk,$req_len,$cmap,$r,$g,$b) = unpack "C$s$l$s$s$s", $request;

print "        ColorMap: " . sprintf "%8.8X", $cmap;
$cmap = Remap_ID($cmap);
print " --> " . sprintf "%8.8X\n", $cmap;

#print "        R/G/B values: " . sprintf("%4.4X %4.4X %4.4X\n", $r,$g,$b);

ReadServer(32);

print $XSERVER pack "CC$s", 127,0,1
    if $main::replay;
}


#############################################################################
# 85
#
# - Ignored

sub X_AllocNamedColor {

my ($request,$cname) = (0,0);
read $CLIENT, $request, 11;
my($junk,$req_len,$cmap,$nam_len,$junk2) = unpack "C$s$l$s$s", $request;
read $CLIENT, $cname, $nam_len;
read $CLIENT, $junk, 4*($req_len-3)-$nam_len;

print "        ColorMap: " . sprintf "%8.8X", $cmap;
$cmap = Remap_ID($cmap);
print " --> " . sprintf "%8.8X\n", $cmap;
print "        ColorName: $cname\n";

ReadServer(32);

print $XSERVER pack "CC$s", 127,0,1
    if $main::replay;
}


#############################################################################
# 91
#
# Bugs: Not yet processing reply
#       - Heck, not even SENDING request now!
#       - Ignored

sub X_QueryColors {

my ($request,$pixels) = (0,0);
read $CLIENT, $request, 7;
my($junk, $req_len, $cmap) = unpack "C$s$l", $request;
read $CLIENT, $pixels, 4*($req_len-2);
  
print "        ColorMap: " . sprintf "%8.8X", $cmap;
$cmap = Remap_ID($cmap);
print " --> " . sprintf "%8.8X\n", $cmap;
print "        Length: $req_len\n"; 

my $reply = 0;
$reply = ReadServer(32);
my($rep,$j1,$seq,$rep_len,$num_rgb) = unpack "CC$s$l$s", $reply;
read $SERVER, $reply, 4*$rep_len;

#print sprintf "%8.8X   %8.8X\n", unpack "$l$l", $pixels;

#print $XSERVER pack "CC$s$l", 91,0,$req_len,$cmap;
#print $XSERVER $pixels;

print $XSERVER pack "CC$s", 127,0,1
    if $main::replay;

} 


#############################################################################
# 93
sub X_CreateCursor {

my $request = 0;
read $CLIENT, $request, 15;
my($junk, $req_len, $cid, $srcpm, $maskpm) = unpack "C$s$l$l$l", $request;
read $CLIENT, $request, 16;

print "        CursorID:     " . sprintf "%8.8X --> ", $cid;
$cid = &Remap_ID($cid);
print sprintf "%8.8X\n", $cid;
print "        SourcePixmap: " . sprintf "%8.8X --> ", $srcpm;
$srcpm = &Remap_ID($srcpm);
print sprintf "%8.8X\n", $srcpm;
print "        Mask Pixmap:  " . sprintf "%8.8X --> ", $maskpm;
$maskpm = &Remap_ID($maskpm);
print sprintf "%8.8X\n", $maskpm;

print $XSERVER pack "CC$s$l$l$l", 93,0,$req_len, $cid, $srcpm, $maskpm
    if $main::replay;
print $XSERVER $request
    if $main::replay;

}


#############################################################################
# 94
sub X_CreateGlyphCursor {

my $request = 0;
read $CLIENT, $request, 15;
my($junk, $req_len, $cid, $source_fnt, $mask_fnt) = 
    unpack "C$s$l$l$l", $request;
read $CLIENT, $request, 16;
print "        CursorID:     " . sprintf "%8.8X --> ", $cid;
$cid = &Remap_ID($cid);
print sprintf "%8.8X\n", $cid;
print "        FontSourceID: " . sprintf "%8.8X --> ", $source_fnt;
$source_fnt = &Remap_ID($source_fnt);
print sprintf "%8.8X\n", $source_fnt;
print "        FontMaskID:   " . sprintf "%8.8X --> ", $mask_fnt;
$mask_fnt = &Remap_ID($mask_fnt);
print sprintf "%8.8X\n", $mask_fnt;

print $XSERVER pack "CC$s$l$l$l", 94,0,$req_len,$cid,$source_fnt,$mask_fnt
    if $main::replay;
print $XSERVER $request
    if $main::replay;

}


#############################################################################
# 96
sub X_RecolorCursor {

my $request = 0;
read $CLIENT, $request, 19;
my($junk,$req_len,$cid,$fr,$fg,$fb,$br,$bg,$bb) =
    unpack "C$s$l$s$s$s$s$s$s", $request;
print "        CursorID: " . sprintf "%8.8X", $cid;
$cid = Remap_ID($cid);
print " --> " . sprintf "%8.8X\n", $cid;

print $XSERVER pack "CC$s$l$s$s$s$s$s$s", 96,0,$req_len,$cid,
               $fr,$fg,$fb,$br,$bg,$bb
		   if $main::replay;
}


#############################################################################
# 98
sub X_QueryExtension {
    my($junk, $junk2, $junk3, $len, $name, $name_len, $new_present, 
       $old_present, $old_reply, $op, $op2, $real_reply, $rep_code, 
       $req_len, $request, $seq);

    ($request, $name) = (0,0);

    read $CLIENT, $request, 7;
    ($junk, $req_len, $name_len, $junk2) = unpack "C$s$s$s", $request;
    read $CLIENT, $name, $name_len;
    print "        Extension Name: $name\n";
    $junk = "";
    read $CLIENT, $junk, 4*($req_len-2)-$name_len;

    #see what the recorded reply was
    $old_reply = ReadServer(32);
    ($rep_code, $junk3, $seq, $len, $old_present,$op) = unpack "CC$s$l"."CC", $old_reply;
    if($rep_code != 1) {die "       ERROR: Recorded reply invalid!\n";}
    print "        Recorded Opcode: " . sprintf "%2.2X ($op decimal)\n", $op 
	if($old_present);
    
    if ($main::replay) {
	print $XSERVER pack "CC$s$s$s", 98,0,$req_len,$name_len,0;
	print $XSERVER $name;
	print $XSERVER $junk;

	#check the real reply, to make sure the real server has it too
	$real_reply = ReadXServer(32);
	($rep_code, $junk, $seq, $len, $new_present,$op2) = 
	    unpack "CC$s$l"."CC", $real_reply;
	if($rep_code != 1) {die "       ERROR: XSERVER's reply invalid!\n";}
	print "        Xserver Opcode:  " . sprintf "%2.2X ($op2 decimal)\n", $op2 
	    if($new_present);
    } else {
	$new_present = $old_present;
    }

    if($old_present != $new_present) {
	if($old_present) {
	    warn "        WARNING: Xserver does not support the expected extension!\n"; 
	}
	else {
	    print "        NOTICE: The recorded server did not support this \n";
	    print "                extension, but the current server does.\n";
	}
    }
    else {
	print "        [Extension supported by both servers.]\n";
    }
}


#############################################################################
# 101
sub X_GetKeyboardMapping {

my $request = 0;
read $CLIENT, $request, 7;
my($junk,$req_len,$key_one,$count) = unpack "C$s"."CC", $request;
print "        Requested $count keycodes starting with $key_one\n";

my $reply = 0;
$reply = ReadServer(32);
my($rep,$kpk,$seq,$rep_len) = unpack "CC$s$l", $reply;
print "        Keysyms per code: $kpk\n";

#grab the keysyms
my $keycode = $key_one;
for(;$keycode<$count+$key_one;$keycode++) {
    read $SERVER, $reply, 4*$kpk;
    $KeyMap[$keycode] = [unpack "L$kpk", $reply]; 
    #print "$keycode: $KeyMap[$keycode][0] $KeyMap[$keycode][1] $KeyMap[$keycode][2] $KeyMap[$keycode][3]\n";
    }

if($HaveKeyMap == 0 ) {
    my($key,$state) = (0,0);
    print "Flushing " . scalar(@KeyQueue)/2 . " Keystrokes...\n";
    while(scalar @KeyQueue) {
        #we saved some keystrokes while waiting for the table...
        $key = shift @KeyQueue;
        $state = shift @KeyQueue;
        XKeyPress($key, $state);
        }
    }
$HaveKeyMap = 1;

print $XSERVER pack "CC$s", 127,0,1
    if $main::replay;
}


#############################################################################
# 104
sub X_Bell {

#this request is ignored

my $request = 0;
read $CLIENT, $request, 3;

print $XSERVER pack "CC$s", 127,0,1
    if $main::replay;
}


#############################################################################
# 119
sub X_GetModifierMapping {

#this request is ignored

my $request = 0;
read $CLIENT, $request, 3;
my $reply = 0;
$reply = ReadServer(32);
my($rep,$kpm,$seq,$rep_len) = unpack "CC$s$l", $reply;
read $SERVER, $reply, 8*$kpm;

print $XSERVER pack "CC$s", 127,0,1
    if $main::replay;

}




#############################################################################
#############################################################################
#############################################################################

sub ReadXServer {

    my $bytes = shift;
    my $expected = shift || 1;
    my ($type, $data) = (0,0);

    if(eof($XSERVER)) {
	print "ERROR: END OF FILE for Xserver!\n";
	die;
    }

    read $XSERVER, $type, 1;
    $type = unpack "C", $type;

    while($type != $expected) {
	print "=== Flushing an Xserver event ($type).........................\n";
	read $XSERVER, $type, 31;
	read $XSERVER, $type, 1;
	$type = unpack "C", $type;
    }

    $bytes--; #oops, we just read one.
    read $XSERVER, $data, $bytes;
    $data = join "", pack("C", $type), $data;

    return $data;
}


#############################################################################
#
# Bugs: If the caller is expecting to read a KeyPress event, the key is
#       not logged. I don't know why this would every occur, though.
sub ReadServer {

my $bytes = shift;
my $expected = shift || 1;      
my ($type, $header, $data) = (0,0,0);

if(eof($SERVER)) {
    print "ERROR: END OF FILE for Server!\n";
    DumpFH($SERVER);
    die;
    }

$len = read $SERVER, $type, 1;
$type = unpack "C", $type; 
    
while($type != $expected) {
    print "=== Flushing a recorded Xserver event ($type)...................\n";
    read $SERVER, $data, 31;
    if($type == 2) { #The server sent a KeyPress event, log it
        my($det,$state,$time,$rid,$eid,$cid,$rx,$ry,$ex,$ey,$state) =
            unpack "C$s$l$l$l$l$s$s$s$s$s", $data;
        if(!$HaveKeyMap) {
            print "        (Buffering keypress code $det)\n";
            push @KeyQueue, ($det, $state);
            }
        else {
            XKeyPress($det, $state);
            }
        }
    last if($type == 0);
    last if(eof($SERVER));
    read $SERVER, $type, 1;
    $type = unpack "C", $type;
    }

#if we're returning a reply, make sure it's the right one
#this is useful when protocol extensions genetate unexpected replies (type 1)
if($type == 1) {
    read $SERVER, $header, 3;
    my($junk, $seq_num) = unpack "C$s", $header;
    if($seq_num < $Req_Seq) {
        print "NOTICE:  Flushing a reply with a unexpected sequence number!\n";
        print "         (Wanted $Req_Seq but got $seq_num)\n";
        #flush the bogus reply
        read $SERVER, $data, 28; #finish reading the standard 32 bytes
        $data = unpack "$l", $data; #get length of any extended fields
        read $SERVER, $data, 4*$data; #flush them too
        $data = ReadServer($bytes,$expected); #a little recursion...
        }
    else {
        print "******** UHOH ******** High seq num!!\n" if($seq_num < $Req_Seq);
        $bytes-=4;
        read $SERVER, $data, $bytes;
        $data = join "", pack("C", $type),$header,$data;
        }
    }
else {
    $bytes--; #oops, we just read one.
    read $SERVER, $data, $bytes;
    $data = join "", pack("C", $type), $data;
    }

return $data;
}


#############################################################################

sub DumpFH {

my $fh  = shift;

print "Test\n";
if(eof($fh)) {
    print STDOUT "DUMPFH: Nothing to do!\n";
    }
else {
print "NO EOF\n\n\n";
    my ($byte, $ascii, $looper)=(0,0,0);
    while(!eof($fh)) {
        if(!$looper) {
            sleep 1;
            print "DUMPFH:  ";
            $ascii = "";
            }
        read $fh, $byte, 1;
        $byte = unpack "C", $byte;
        print sprintf "%2.2X  ", $byte;
        if(($byte > 31) && ($byte < 127)) {
            $ascii = join '', $ascii, pack "C", $byte;
            }
        else {
            $ascii = join '', $ascii, '.';
            }
        $looper = ($looper+1)%8;
        print "  $ascii\n" if(!$looper);
        }
    print "\n";
    }
}


#############################################################################
#############################################################################

my ($new_rid_base, $new_rid_mask, $old_rid_base, $old_rid_mask);
my ($new_root_id, $old_root_id, $new_root_cmap, $old_root_cmap);
my(%server_pixpad, %xserver_pixpad);
my (@X_Request, @X_Request_Name);
my ($l, $s, $CLIENT, $SERVER, $XSERVER, $KEYS);
my (@SymMap, @Keymap, @KeyQueue, $HaveKeyMap);
my (%atom_map);

1;
